#!@BIN_PERL@
#
#
#    Short description: Creates/Updates RRA archives, from nagios performance data
#    Copyright (C) 2006 - 2007 Badri Prasad Pillai
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

#
#    Contact details:
#               Badri(at)DigLinks.COM

#
$| = 1;    # disable buffering

use strict;

# Variable declaration
use vars qw(
  $query_string   @key_value
  %info        %graph           $cache_dir
  $rrd_image_file   $rrd_file    @rrdtool_options @rrdtool_source
  $graph_template %config      @zoom_javascript $graph_height
  $graph_width    $total_graphs   $version      $footer
  $rrd_files      $tmp_rrd_files    $conf_file 	$data_flow      $debug
  $col            $g 		$RRD_ERROR %query_variables %query_variables_have
  $nagios_env_file $nagios_env $nagios_env_count $ds_menu @dsnames $dshash
);

######################################################
# Change following variable if required
######################################################
$conf_file = "/etc/n2rrd/n2rrd.conf";

#
# commnet following line if you want to use configuration file
# to overide some default variables
#
#undef $conf_file;
undef $RRD_ERROR;

$debug = 1;

#
# if $DATA_FLOW$ variable is defined in graph template
# then add the daily,weekly,monthely and yearly seconds value to calculate e.g. throughput
#
$data_flow = 0;

######################################################
# Normaly no need to change any things below this line
######################################################

$version = "1.3.2";
$footer =
" <br><strong>Query URL: http://$ENV{SERVER_NAME}$ENV{SCRIPT_NAME}?$ENV{QUERY_STRING} </strong> <HR>Generated by <a href=\"http://n2rrd.diglinks.com\""
  . ">rrd2graph</a>($version), total graphs =";

$rrd_files        = {};
$tmp_rrd_files    = {};
%info             = ();
@rrdtool_options  = ();
$total_graphs     = 0;
$nagios_env_count = 0;
sub wanted;
sub create_thumb ($$$$$);
sub display_host ($$);
sub summary($);

sub log_msg($);
sub zoom ($$);
sub load_config($);
sub load_nagios_env($);
sub list_default_values();

#
# If no configuration file given, then use default values
#
%config = (
    CONF_DIR                     => "/etc/n2rrd",
    TEMPLATES_DIR                => "templates",
    SERVICE_NAME_MAPS            => "templates/maps/service_name_maps",
    ZOOM_JS                      => "js/zoom.js",
    RRA_DIR                      => "/var/log/monitor/n2rrd/rra",
    LOGFILE                      => "/var/log/monitor/n2rrd/rra/n2rrd.log",
    DOCUMENT_ROOT                => "/var/www/html",
    CACHE_DIR                    => "rrd-images",
    CGIBIN                       => "perl",
    NAGIOS_CGIBIN                => "nagios/cgi-bin",
    RRDTOOL                      => "/usr/bin/rrdtool",
    RRD_PATH_HIDDEN              => 0,
    NAGIOS_HOST_URL              => 1,
    TMPDIR                       => "/tmp",
    THUMB_WIDTH                  => 200,
    THUMB_HEIGHT                 => 100,
    THUMB_DISPLAY                => "Daily",
    THUMB_DISPLAY_COLUMNS        => 3,
    SAVE_NAGIOS_ENVIRONMENT_VARS => 0,
    TEMPLATE_NAME_AT_BEGINNING   => 0,
);

%query_variables = (
    service         => "",
    match_host      => "",
    summary         => "",
    hostname        => "",
    action          => "",
    display         => "",
    graph_height    => "",
    graph_width     => "",
    graph_start     => "",
    graph_end       => "",
    title_font_size => "",
    el              => "",
    source          => "",
    extmacro        => "",
    extvalue        => "",
    ds_name         => ""
);

%query_variables_have = ();

load_config($conf_file) if $conf_file;

#
# Abort if query string is not defined
log_msg "Environment Variable QUERY_STRING undefined\n"
  and list_default_values
  if not $ENV{'QUERY_STRING'};


#
# set to rrd default
$graph_width  = 400;
$graph_height = 100;

@key_value = ();

%graph = (
    Daily           => "--start=-1d",
    Daily_seconds   => 86400,
    Weekly          => '--start=-604800',
    Weekly_seconds  => 604800,
    Monthly         => '--start=-1m',
    Monthly_seconds => 2592000,
    Yearly          => '--start=-1y',
    Yearly_seconds  => 31536000
);

$query_string = $ENV{'QUERY_STRING'};

# save key value pair
@key_value = split( /&/, $query_string );

#
# evaluate each value
foreach my $kv (@key_value) {

    # MIME code cleaning
    $kv =~ tr/+/ /;

    # MIME 8-bit hex  value convertion
    $kv =~ s/%([a-fA-F0-9].)/pack("C", hex($1))/eg;

    # XSSI clean !!
    $kv =~ s/<!--(.|\n)*-->//g;

    # Remove unwanted characters
    $kv =~ s/[^a-zA-Z0-9\-_=\.:\ ]//g;

    my ( $key, $value ) = split( /=/, $kv );

    $config{SAVE_NAGIOS_ENVIRONMENT_VARS} = $value if $key eq "extended_info";
    next if $key eq "extended_info";
    next if $key eq "ds_name" and $value eq "undefined";

    #
    # Validate variable name
    #
    if ( defined $query_variables{$key} ) {
        if ( length($value) > 0 ) {
            if ( defined $query_variables_have{$key} ) {
                log_msg "Query string variable name \"$key\" already defined\n";
                exit 1;
            }
            else {
                $info{$key} = $value;
                $query_variables_have{$key}++;
            }
        }
        else {
            log_msg "Missing value for query string variable name \"$key\"\n";
            next;
        }
    }
    else {
        log_msg "Unknown query string variable name \"$key\"\n";
        exit 1;
    }

    #print "$key = $value\n";
}

#
# assume there are no nagios variables if NAGIOS_URL switch is disabled
#
$config{SAVE_NAGIOS_ENVIRONMENT_VARS} = 0 if not $config{NAGIOS_HOST_URL};

if ( !defined $info{summary} ) {
    @rrdtool_source  = ();
    @rrdtool_options = ();

    #
    # Sanity check: if RRA directory exists
    log_msg "Direcotry rra_dir = \"$config{CONF_DIR}\" doesn't exists\n"
      and exit 1
      if !-d $config{CONF_DIR};

    #
    # Sanity check: if templates directory exists
    log_msg
"Direcotry templates_dir = \"$config{CONF_DIR}/$config{TEMPLATES_DIR}\" doesn't exists\n"
      and exit 1
      if !-d "$config{CONF_DIR}/$config{TEMPLATES_DIR}";

    #
    # save original service name
    $info{service_orig} = $info{service};

    #
    # assume delimiter between service name is '_'
    my @parse_service_str = split /_/, $info{service};

    if ( $#parse_service_str <= 0 ) {
        open S_MAPS, "< $config{CONF_DIR}/$config{SERVICE_NAME_MAPS}"
          or log_msg
"Can't open service maps file \"$config{CONF_DIR}/$config{SERVICE_NAME_MAPS}\"\n"
          and exit 1;

        log_msg
"Searching map in file \"$config{CONF_DIR}/$config{SERVICE_NAME_MAPS}\" for service \"$info{service}\" \n"
          if $debug;

        while (<S_MAPS>) {
            next if /^#/;    # Skip comments
            next if /^$/;    # Skip empty lines
            s/#.*//;         # Remove partial comments
            chomp;

            if (/^$info{service}:\s+(\S+)/i) {
                $info{service} = $1 if /^$info{service}:\s+(\S+)/i;
                log_msg "Found service map -> $info{service}\n" if $debug;
            }
        }
    }
    else {
        if ( $config{TEMPLATE_NAME_AT_BEGINNING} ) {

            # The first word after terminator character "_"
            $info{service} = $parse_service_str[0];
        }
        else {

            # The last word after terminator character "_"
            $info{service} = $parse_service_str[$#parse_service_str];
        }
    }

    #
    # from config info rrd_file is the following
    $rrd_file =
      "$config{RRA_DIR}/$info{hostname}/$info{hostname}_$info{service}.rrd";

    #
    # check order same as rra template
    # 1. hostname_service.t
    # 2. service.t
    # 3. as per rewrite info
    if ( defined( $info{ds_name} ) ) {
        if (
            -f "$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/$info{hostname}_$info{service}_$info{ds_name}.t"
          )
        {
            $graph_template =
"$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/$info{hostname}_$info{service}_$info{ds_name}.t";
        }
        elsif (
            -f "$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/$info{service}_$info{ds_name}.t"
          )
        {
            $graph_template =
"$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/$info{service}_$info{ds_name}.t";
        }
        elsif (
            -f "$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/default_ds_name.t"
          )
        {
            $graph_template =
"$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/default_ds_name.t";
        }
    }
    else {
        $graph_template =
"$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/$info{hostname}_$info{service}.t";
        unless ( -e $graph_template ) {
            $graph_template =
              "$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/$info{service}.t";
        }
    }

    # Check if a rewrite rules exist
    #
    if (
        -f "$config{CONF_DIR}/$config{TEMPLATES_DIR}/rewrite/$info{hostname}_$info{service}_rewrite"
      )
    {

        #
        #	print "Rewrite file detected for: ${hostname}_${service}\n";
        open REWRITE,
"<$config{CONF_DIR}/$config{TEMPLATES_DIR}/rewrite/$info{hostname}_$info{service}_rewrite"
          or log_msg "Can't open rewrite rules file for $info{hostname}\n"
          and exit 1;

        while (<REWRITE>) {
            next if /^#/;    # Skip comments
            next if /^$/;    # Skip empty lines
            s/#.*//;         # Remove partial comments
            chomp;

            $rrd_file = $1 if /^rrd_file\s+(\S+)/;

            #
            # graph template
            $graph_template = $1 if /^graph_template\s+(\S+)/;
        }
    }

    if ( defined $info{el} and $info{el} == 1 ) {
        if ( not defined( $info{ds_name} ) ) {
            $graph_template =
"$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/$info{hostname}_$info{service}_el.t";
            unless ( -e $graph_template ) {
                $graph_template =
"$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/default_execution_latency_el.t";
            }
        }

        $rrd_file =~ s/\.rrd$/_el\.rrd/;
        if ( !-f $rrd_file ) {
            $rrd_file =
"$config{RRA_DIR}/$info{hostname}/$info{hostname}_$info{service}_el.rrd";
        }
    }

    log_msg
"RRD file \"$rrd_file\" missing for host=\"$info{hostname}\" service=\"$info{service}\"\n"
      and exit 1
      if !-f $rrd_file;

    open G, "<$graph_template"
      or log_msg "can't open graph template \"$graph_template\"\n" and exit 1;

    while (<G>) {
        next if /^#/;    # Skip comments
        next if /^$/;    # Skip empty lines

        chomp;
        s/^\s+//;
        s/\s+$//;
        s/\$HOSTNAME\$/$info{hostname}/g        if /\$HOSTNAME\$/;
        s/\$SERVICENAME\$/$info{service_orig}/g if /\$SERVICENAME\$/;
        s/\$RRD_FILENAME\$/$rrd_file/g          if /\$RRD_FILENAME\$/;
        s/\$DSNAME\$/$info{ds_name}/g           if /\$DSNAME\$/;

        #
        # Store graph height and width for later use
        #
        $graph_height = $1 if /(^-h|^--height)/ and m/(\d+)/;
        $graph_width  = $1 if /(^-w|^--width)/  and m/(\d+)/;

        #
        # looks like ethernet bandwidth calculation is enabled
        #
        $data_flow = 1 if /\$DATA_FLOW\$/;

        push @rrdtool_options, $_;

        s#$config{RRA_DIR}#PATH_HIDDEN# if $config{RRD_PATH_HIDDEN};

        # in case rrd file name is hard coded in template
        s#=.*/#=PATH_HIDDEN/# if /\.rrd/ and $config{RRD_PATH_HIDDEN};

        push @rrdtool_source, $_ if defined $info{source};
    }

    close G;

    my $now = localtime(time);
    $now =~ s/:/\\:/g;

    push @rrdtool_options, "COMMENT:\\n";
    push @rrdtool_options, "COMMENT:Create date\\: $now";
}
else {
    $config{SAVE_NAGIOS_ENVIRONMENT_VARS} = 0;
}

$col = 1;

if ( !defined $info{action} or ( $info{action} ne "image" ) ) {
    print "Content-Type: text/html\r\n\r\n";

    print "<head>\n\n";
    print "<title>\n\n";
    if ( defined $info{summary} ) {
        print "rrd2graph - summary \n\n";
    }
    else {
        print
"rrd2graph - for host $info{hostname} and service $info{service} \n\n";
        $cache_dir =
          "$config{DOCUMENT_ROOT}/$config{CACHE_DIR}/$info{hostname}";

        #
        # Creat hostname directory if necessary
        mkdir "$cache_dir" if !-d "$cache_dir";
    }
    print "</title>\n\n";
    print
"<meta http-equiv=refresh content='300'; url=\"/$config{CGIBIN}/rrd2graph.cgi?$ENV{QUERY_STRING}\">\n";
    print "<STYLE TYPE=\"text/css\">\n";
    print "<!--  \n";
    print
"BODY, TABLE, TR, TD {font-family: Verdana, Arial, Helvetica, sans-serif; font-size: 10px;}\n";
    print "BODY {background-color: #FFFFFF;}\n";
    print "A {text-decoration: none;}\n";
    print "A:active { text-decoration: none;}\n";
    print "A:hover {text-decoration: underline; color: #333333;}\n";
    print "A:visited {color: Blue;}\n";
    print ".textHeader {font-size: 12px; font-weight: bold;}\n";
    print
      ".textHeaderDark {font-size: 12px; font-weight: bold; color: #ffffff;}\n";
    print ".textArea {font-size: 12px;}\n";
    print "-->\n";
    print "</STYLE>\n";

    print <<H_JS
    
    <script language="JavaScript">
    <!--
    function help(Click_Menu)
    // Function that will swap the display/no display for
    // all content within span tags
    {
        if (Click_Menu.style.display == "none") {
            Click_Menu.style.display = "";
        }
        else {
           Click_Menu.style.display = "none";
        }
    }
    -->
    </script>
H_JS
      ;
    print "</head>\n\n";
    print "<html>\n\n";
    print '<body leftmargin="0" topmargin="0" marginwidth="0" marginheight="0">'
      . "\n\n";

    #
    # possibility to search by service/ds/host combination
    if ( defined $info{summary} ) {
    my $h_search = <<H_Search;
    <span ID="h_search" style="display: none">
     <table bgcolor="#FFFFCC">
        <tr>
          <td width=300 wrap>
            <strong>
             Service:
            </strong> this field expects service name or substr of service name <br>
            <strong>
             DS Name:
            </strong> this field expects exact DS (Data Source) Name, works only if Service is defined<br>
            <strong>
             Match Host:
            </strong> this field expects host name or substr of host name<br>
          </td>
        </tr>
	  <td align='center'>
            <a href=\"javascript:help(h_search)\">Close</a>
	  </td>
     </table>
    </span>
H_Search

        print
"<table align='center'><tr><td align='center' colspan=$config{THUMB_DISPLAY_COLUMNS}>";
        print
          "    <form  name=\"input\" action=\"/$config{CGIBIN}/rrd2graph.cgi\"
                    method=\"get\"> 
                      </strong><input type=\"hidden\" name=\"summary\" value=\"all\" 
              ";

        print
"     <strong>Service: </strong> <input type=\"text\" name=\"service\" value=\"$info{service}\">\n";
        print
"     <strong>DS name: </strong> <input type=\"text\" name=\"ds_name\" value=\"$info{ds_name}\">\n";
        print "     <input type=\"hidden\" name=\"el\"      value=\"1\">\n"
          if $info{el};
        print " <strong>Match host: </strong>
            <input type=\"text\" name=\"match_host\"value=\""
          . ( defined $info{match_host} ? "$info{match_host}\">" : "\">" )
          if $info{summary};
        print " <SELECT NAME=\"display\"> ";

        foreach my $dg ( keys %graph ) {
            next if $dg =~ /seconds/;
            if ( $dg eq $config{THUMB_DISPLAY} ) {
                print "<OPTION VALUE=\"$dg\" SELECTED> $dg  ";
            }
            else {
                print "<OPTION VALUE=\"$dg\"> $dg";
            }
        }
        print "
                    </SELECT>
                    <input type=\"submit\" value=\"Submit\">
            <a href=\"javascript:help(h_search)\">Help</a> $h_search</form></td></tr></table>";
    }

    print "<table align='center' width='1' cellpadding='0'>\n";
    print "<tr><td>\n" if $config{SAVE_NAGIOS_ENVIRONMENT_VARS};
}
$g = "";

if ( defined $info{summary} ) {

    use File::Find ();

    #
    # Summary related
    #
    File::Find::find( { wanted => \&wanted }, $config{RRA_DIR} );
    summary $info{summary};
}
else {

    if ( $config{SAVE_NAGIOS_ENVIRONMENT_VARS} ) {
        $nagios_env      = "";
        $nagios_env_file = $rrd_file;
        $nagios_env_file =~ s/(_el\.rrd|\.rrd)//;
        $nagios_env_file .= ".env";

        if ( -f $nagios_env_file ) {
            load_nagios_env($nagios_env_file);
            print "<table><tr><td>\n" if not defined $info{action};
        }
        else {
            log_msg
              "Nagios environment vars file \"$nagios_env_file\" missing\n";
            $config{SAVE_NAGIOS_ENVIRONMENT_VARS} = 0;
        }
    }

    use RRDs;

    #
    # Extract ds names from rrd_file
    $dshash = RRDs::info "$rrd_file";

    @dsnames = ();
    foreach my $key ( sort keys %$dshash ) {
        push @dsnames, $1 if $key =~ m/^ds\[(.*)\]\.type/;
    }

    foreach my $d ( "Daily", "Weekly", "Monthly", "Yearly" ) {
        $g = $d;

        $g = $info{display}
          if defined $info{display}
              and defined $graph{ $info{display} };

        $total_graphs++;

        if ( defined $info{action} ) {
            my ( $rrd_start, $step, $names, $data ) =
              RRDs::fetch( "$rrd_file", "AVERAGE", "-s",
                $graph{ $g . "_seconds" } * -1 );

            if ( defined $info{graph_end} ) {
                my $now = time;

                $info{graph_end} = $now - $step
                  if $info{graph_end} > ( $now - $step );

          #$info{graph_start} = $rrd_start  if  $info{graph_start} < $rrd_start;

                # required for zoom out function
                $info{graph_start}-- if $info{graph_start} == $info{graph_end};

            }
            $info{graph_start} = $rrd_start if !defined $info{graph_start};
            $info{graph_end} = $rrd_start + ( ($#$data) * $step )
              if !defined $info{graph_end};

            zoom( $graph{ $g . "_seconds" }, $rrd_file );
            last;
        }

        my $rrd_image_file =
          "$cache_dir/$info{hostname}_$info{service}_${g}.png";

        if ( $info{ds_name} ) {
            $rrd_image_file =
"$cache_dir/$info{hostname}_$info{service}_$info{ds_name}_${g}.png"
              if not $info{el};
            $rrd_image_file =
"$cache_dir/$info{hostname}_$info{service}_$info{ds_name}_${g}_el.png"
              if $info{el};
        }
        else {
            $rrd_image_file =~ s/\.png$/_el\.png/ if $info{el};
        }

        foreach (@rrdtool_options) {
            s/"//g;
            s/\$DATA_FLOW\$/$graph{$g . "_seconds"}/g if $data_flow;
        }

        my ( $a, $xsize, $ysize ) =
          RRDs::graph( $rrd_image_file, $graph{$g}, @rrdtool_options );

        $RRD_ERROR = RRDs::error;
        log_msg
          "ERROR \"hostname = $info{hostname} -> service = $info{service}\"\
				 graph failed : $RRD_ERROR\n" if $RRD_ERROR;

        if ( $config{SAVE_NAGIOS_ENVIRONMENT_VARS} ) {
            print
"<tr> <td align='left'><a href=\"/$config{CGIBIN}/rrd2graph.cgi?$ENV{QUERY_STRING}&extended_info=0\">Extended Information ON</a></td></tr>\n"
              if $total_graphs == 1
                  and $config{NAGIOS_HOST_URL};
        }
        else {
            print
"<tr> <td align='left'><a href=\"/$config{CGIBIN}/rrd2graph.cgi?$ENV{QUERY_STRING}&extended_info=1\">Extended Information OFF</a></td></tr>\n"
              if $total_graphs == 1
                  and $config{NAGIOS_HOST_URL};
        }

        #
        # Diaplay source
        #
        if ( defined $info{source} ) {

            #print "<tr> <td align='left'>\n";
            print "<tr> <td valign=\"top\" bgcolor=\"#FFFFCC\">\n";
            $rrd_image_file =~ s#$cache_dir#PATH_HIDDEN#
              if $config{RRD_PATH_HIDDEN};
            print
"<PRE> $config{RRDTOOL} graph $rrd_image_file \\\n $graph{$g} \\\n";
            foreach (@rrdtool_source) {
                s/\$DATA_FLOW\$/$graph{$g . "_seconds"}/g if $data_flow;

                print;
                print " \\\n";
            }
            print "</PRE> </td> </tr>\n";
        }

        #
        # Well any better solution to get the "Select DS" menu output ordered
        # is welcome
        print "<tr> <td align='center'>\n" if $#dsnames <= 0;
        print "<tr> <td align='right'>\n"  if $#dsnames > 0;

        if ( $info{el} ) {
            print
"<img src=\"/$config{CACHE_DIR}/$info{hostname}/$info{hostname}_$info{service}"
              . (
                $info{ds_name}
                ? "_$info{ds_name}_${g}_el.png\" "
                : "_${g}_el.png\" "
              )
              . "alt=\"Plugin Execution and Latency graph image\"> <br> "
              . "[<strong><a href=\"/$config{CGIBIN}/rrd2graph.cgi?hostname=$info{hostname}&"
              . "service=$info{service}&el=1&display=$g\">$g</a></strong>] \n";
            print
"[<a href=\"/$config{CGIBIN}/rrd2graph.cgi?hostname=$info{hostname}\&service=$info{service}\">"
              . "Back to \"$info{service}\" performance graph</a>] ";
            print
" [<a href=\"/$config{NAGIOS_CGIBIN}/status.cgi?navbarsearch=1&host=$info{hostname}\">Nagios</a>] "
              if $config{NAGIOS_HOST_URL};
            print
" [<a href=\"/$config{CGIBIN}/rrd2graph.cgi?hostname=$info{hostname}&service=$info{service}&"
              . "el=1&display=$g"
              . ( $info{ds_name}
                ? "&ds_name=$info{ds_name}&source=1"
                : "&source=1" )
              . "\">Source</a>]"
              . " [<a href=\"/$config{CGIBIN}/rrd2graph.cgi?action=zoom&hostname=$info{hostname}"
              . "&service=$info{service}&display=$g&graph_height=$graph_height&graph_width=$graph_width&el=1"
              . ( $info{ds_name} ? "&ds_name=$info{ds_name}\"" : "\"" )
              . ">Zoom</a>]";
        }
        else {
            print
"<img src=\"/$config{CACHE_DIR}/$info{hostname}/$info{hostname}_$info{service}"
              . (
                $info{ds_name} ? "_$info{ds_name}_${g}.png\"" : "_${g}.png\"" )
              . " alt=\"$g graph image\"> <br> <strong>[<strong><a href=\"/$config{CGIBIN}/rrd2graph.cgi?"
              . "hostname=$info{hostname}\&service=$info{service}&display=$g\">$g</a></strong>] \n";
            if (
"$config{DOCUMENT_ROOT}/$config{CACHE_DIR}/$info{hostname}/$info{hostname}_$info{service}_${g}_el.png"
              )
            {
                print
"[<a href=\"/$config{CGIBIN}/rrd2graph.cgi?hostname=$info{hostname}\&service=$info{service}"
                  . "&el=1\">Plugin Execution and Latency Time</a>] ";
                print
"[<a href=\"/$config{NAGIOS_CGIBIN}/status.cgi?navbarsearch=1&host=$info{hostname}\">Nagios</a>] "
                  if $config{NAGIOS_HOST_URL};
                print
" [<a href=\"/$config{CGIBIN}/rrd2graph.cgi?hostname=$info{hostname}&service=$info{service}"
                  . "&display=$g"
                  . ( $info{ds_name}
                    ? "&ds_name=$info{ds_name}&source=1"
                    : "&source=1" )
                  . "\">Source</a>]"
                  . " [<a href=\"/$config{CGIBIN}/rrd2graph.cgi?action=zoom&hostname=$info{hostname}"
                  . "&service=$info{service}&display=$g&graph_height=$graph_height&graph_width=$graph_width&el=0"
                  . ( $info{ds_name} ? "&ds_name=$info{ds_name}\"" : "\"" )
                  . ">Zoom</a>]";
            }
        }
        print
" [Graphs <a href=\"/$config{CGIBIN}/rrd2graph.cgi?summary=$info{hostname}\">Host</a>"
          . "|<a href=\"/$config{CGIBIN}/rrd2graph.cgi?summary=all\">All</a>]";

        #
        # If more than one DS in rrd_file then display "Select DS" menu
        if ( $#dsnames > 0 ) {
            print
" [<a href=\"javascript:help(ds_menu_$g)\">Select DS</a>]<br><br>";
            $ds_menu = "<span ID=\"ds_menu_$g\" style=\"display: none\">"
              . "<table bgcolor=\"#FFFFCC\">";

            $ds_menu .=
"<tr><td width=25 align=left><strong>$info{service}</strong></td><td>";
            $ds_menu .=
"<a href=\"/$config{CGIBIN}/rrd2graph.cgi?hostname=$info{hostname}&service=$info{service}";
            $ds_menu .= ( $info{el} ? "&el=1" : "" );
            $ds_menu .= "\">$info{hostname}</a>|";
            my $strip_svc = $info{service};
            $strip_svc =~ s/^_//;
            $ds_menu .=
"<a href=\"/$config{CGIBIN}/rrd2graph.cgi?summary=all&service=$strip_svc";
            $ds_menu .= ( $info{el} ? "&el=1" : "" );
            $ds_menu .= "\">All</a> </td> </tr>";

            foreach my $ds (@dsnames) {
                $ds_menu .= "<tr><td width=25 align=left>$ds</td><td>";
                $ds_menu .=
"<a href=\"/$config{CGIBIN}/rrd2graph.cgi?hostname=$info{hostname}&service=$info{service}";
                $ds_menu .= ( $info{el} ? "&el=1" : "" );
                $ds_menu .= "&display=$g&ds_name=$ds\">$g</a>|";
                $ds_menu .=
"<a href=\"/$config{CGIBIN}/rrd2graph.cgi?hostname=$info{hostname}&service=$info{service}";
                $ds_menu .= ( $info{el} ? "&el=1" : "" );
                $ds_menu .= "&ds_name=$ds\">DWMY</a>|";
                $ds_menu .=
"<a href=\"/$config{CGIBIN}/rrd2graph.cgi?summary=all&service=$info{service}";
                $ds_menu .= ( $info{el} ? "&el=1" : "" );
                $ds_menu .= "&ds_name=$ds\">All</a> </td> </tr>";
            }
            $ds_menu .= "</table> </span>";
            print $ds_menu;
        }
        else {
            print "<br><br>\n";
        }
        print "</td> </tr>\n";
        last if defined $info{source};
        last if $info{display} and defined $graph{ $info{display} };
    }
}

if ( !defined $info{graph_end} ) {
    print "</table> \n";
    if ( $config{SAVE_NAGIOS_ENVIRONMENT_VARS} ) {
        print "</td><td valign='top'> $nagios_env </td></tr></table>\n"
          and $nagios_env_count++
          if not $nagios_env_count;
    }
    print "$footer" . $total_graphs;
    print "</body>\n\n";
    print "</html>\n\n";
}

sub log_msg ($) {
    my $str = shift;

    if ($debug) {
        open L, ">>$config{LOGFILE}"
          or die "Can't open logfile \"$config{LOGFILE}\"";
        print L scalar localtime() . ":rrd2graph: $str";
        close L;
    }
    else {
        print "Content-Type: text/html\r\n\r\n";
        print "<head>\n\n";
        print "<title>\n\n";
        print "n2rrd (rrd2graph) \n\n";
        print "</title>\n\n";

        print "</head>\n\n";
        print "<html>\n\n";
        print "<body>\n\n";
        print "<HR> <H2> rrd2graph:ERROR: </H2><BR>\n";
        print "<P align=left>\n";
        print "	<FONT face=\"Comic Sans MS\" size=4> \n";
        print "$str";
        print "<BR><BR>\n</FONT>\n</P>\n ";
        print "$footer" . $total_graphs;
        print "</body>\n\n";
        print "</html>\n\n";
    }
}

sub summary ($) {
    my $op = shift;

    if ( $op eq "all" ) {

        foreach my $h ( sort keys %{$rrd_files} ) {

            my $found_env = 0;
            my $tmp_s     = "";
            $tmp_rrd_files = {};
            if ( ( defined $info{extmacro} ) and ( defined $info{extvalue} ) ) {
                foreach my $s ( sort keys %{ $rrd_files->{$h} } ) {
                    $tmp_s = $s;
                    $s =~ s/rrd/env/;
                    my $env_file = "$config{RRA_DIR}/$h/$h" . '_' . "$s";
                    open( ENV, "<$env_file" ) or next;
                    while (<ENV>) {
                        chomp;
                        my ( $k, $v ) = split /=/, $_, 2;
                        if (    ( $k eq $info{extmacro} )
                            and ( $v eq $info{extvalue} ) )
                        {
                            $found_env = 1;
                            $tmp_rrd_files->{$h}->{$tmp_s}++;
                            last;
                        }
                    }
                    close ENV;

                    #last if $found_env;
                }
                if ($found_env) {
                    display_host $h, \%{$tmp_rrd_files};
                }
                next;
            }
            display_host $h, \%{$rrd_files};
        }
    }
    else {
        if ( defined $rrd_files->{$op} ) {
            display_host $op, \%{$rrd_files};
            if ( $col < 4 ) {
                print "\t<td> </td>\n<td> </td>\n</tr>\n" if $col == 2;
                print "\t<td> </td>\n</tr>\n"             if $col == 3;
            }
        }
        else {
            print "<HR> <H2> rrd2graph:ERROR: </H2><BR>\n";
            print "<P align=left>\n";
            print " <FONT face=\"Comic Sans MS\" size=4> \n";
            print "This host \"$op\" does not exist";
            print "<BR><BR>\n</FONT>\n</P>\n\n";
        }
    }
}

sub display_host ($$) {
    my $host         = shift;
    my $ref          = shift;
    my $use_rrd_file = "";
    my $tt           = "";

    foreach my $s ( sort keys %{ $ref->{$host} } ) {
        my $t = $s;
        $t =~ s/\.rrd//;
        $tt = $t;

        if ( $info{el} ) {
            $s = "$t" . "_el" . ".rrd";
            $t .= "_el";
        }
        else {
            $t .= "_" . $info{ds_name} if $info{ds_name};
        }

        print "<tr>\n" if $col == 1;

        my $graph_t = "$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/${t}.t";
        $graph_t =
"$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/default_execution_latency_el.t"
          if $info{el};
        if ( $info{ds_name} ) {
            if ( !-e "$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/${t}.t" ) {
                $graph_t =
"$config{CONF_DIR}/$config{TEMPLATES_DIR}/graph/default_ds_name.t";
            }
        }

        if ( -e "$graph_t" ) {
            mkdir "$config{DOCUMENT_ROOT}/$config{CACHE_DIR}/$host"
              if !-d "$config{DOCUMENT_ROOT}/$config{CACHE_DIR}/$host";

            create_thumb $host,
              ( $info{ds_name} ? "$tt->$info{ds_name}" : $tt ),
              "$graph_t",
              "$config{RRA_DIR}/$host/${host}_$s",
"$config{DOCUMENT_ROOT}/$config{CACHE_DIR}/$host/${host}_${t}_tb.png";
            if (
                -e "$config{DOCUMENT_ROOT}/$config{CACHE_DIR}/$host/${host}_${t}_tb.png"
              )
            {
                print
"\t<td> <a href=/$config{CGIBIN}/rrd2graph.cgi?hostname=$host&service="
                  . ( $info{ds_name} ? "$tt&ds_name=$info{ds_name}" : "$tt" )
                  . ( $info{el}      ? "&el=1"                      : "" )
                  . ( $info{display} ? "&display=$info{display}>"   : ">" )
                  . "<img src=\"/$config{CACHE_DIR}/$host/${host}_${t}_tb.png\"> </a> </td>\n";
                $col++;
            }

            print "</tr>\n" and $col = 1
              if $col == ( $config{THUMB_DISPLAY_COLUMNS} + 1 );
            $total_graphs++;
        }
    }
}

sub create_thumb ($$$$$) {
    my ( $host, $service, $template, $rrd_file, $thumb_image_filename ) = @_;

    my $rrd_start = $graph{ $config{THUMB_DISPLAY} };
    $rrd_start = $graph{ $info{display} }
      if defined $info{display}
          and defined $graph{ $info{display} };

    use RRDs;

    @rrdtool_options = ();

    push @rrdtool_options, "--no-legend";

    open G_TEMP, "<$template"
      or die "can't open graph template \"$template\"" and exit 1;

    while (<G_TEMP>) {
        next if /^#/;                     # Skip comments
        next if /^$/;                     # Skip empty lines
        next if /-z/ and $info{el};
        next if /--lazy/ and $info{el};
        chomp;

        s/^\s+//;
        s/\s+$//;
        s/"//g;

        s/\$RRD_FILENAME\$/$rrd_file/g if /\$RRD_FILENAME\$/;
        s/\$DSNAME\$/$info{ds_name}/g  if /\$DSNAME\$/;

        #
        # remove width and height, and provide new one one
        $_ = "--width=$config{THUMB_WIDTH}" if /-w\s+/;
        $_ = "--width=$config{THUMB_WIDTH}" if /--width=/;

        $_ = "--height=$config{THUMB_HEIGHT}" if /-h\s+/;
        $_ = "--height=$config{THUMB_HEIGHT}" if /--height=/;

        $_ = "--title=$host - $service" if /-t\s+/;
        $_ = "--title=$host - $service" if /--title=/;

        $_ =~ s/\$DATA_FLOW\$/86400/g if /\$DATA_FLOW\$/;

        push @rrdtool_options, $_;
    }

    close G_TEMP;

    #
    # remove image if default display value changes temporarly
    unlink $thumb_image_filename if $info{Display} ne $config{THUMB_DISPLAY};
    my ( $a, $xsize, $ysize ) =
      RRDs::graph( $thumb_image_filename, $rrd_start, @rrdtool_options );

    $RRD_ERROR = RRDs::error;
    log_msg
"ERROR \"hostname = $host -> service = $service\" graph failed : $RRD_ERROR\n"
      if $RRD_ERROR;

    return 0;
}

sub wanted {
    my ( $dev, $ino, $mode, $nlink, $uid, $gid, $host, $service );

    ( ( $dev, $ino, $mode, $nlink, $uid, $gid ) = lstat($_) );

    #
    # Store hostname and service name
    #
    if ( -f _ ) {
        if ( $_ =~ /_el\.rrd$/s ) {
            return;
        }
        elsif ( $_ =~ /^(.*)_([A-Za-z0-9\-]+\.rrd)$/s ) {
            $host    = $1;
            $service = $2;
        }

        #( $host, $service ) = split '_', $_, 2;
        return if !$service;
        return if $info{service} and $service !~ /$info{service}/;
        return if defined $info{match_host} and $host !~ /$info{match_host}/;
        $rrd_files->{$host}->{$service}++ if -f $_;
    }
}

sub zoom ($$) {
    my ( $start, $rrd_filename ) = @_;

    #
    # for bandwidth calculation
    #
    if ($data_flow) {
        my ( $rrd_start, $step, $names, $data ) =
          RRDs::fetch( "$rrd_filename", "AVERAGE", "-s", $info{graph_start},
            "-e", $info{graph_end} );

        my $end = $rrd_start + ( ($#$data) * $step );
        $start = $end - $rrd_start;
    }

    $info{el} = 0 if $info{el} != 1;

    my $help_text = <<H_TEXT
    <span ID="zoom_help" style="display: none">
     <table bgcolor="#FFFFCC">
        <tr>
          <td width=300 wrap>
            <strong>
             Zoom in: 
            </strong> with "left mouse key" pressed, mark graph area and release key. <br>
            <strong>
             Zoom out:
            </strong> click graph area with "right mouse key"
          </td>
        </tr>
     </table>
    </span>
H_TEXT
      ;

    if ( $info{action} eq "zoom" ) {
        print <<DIV
	<div id='zoomBox' style='position:absolute; overflow:none; left:0px; top:0px; width:0px; height:0px; visibility:visible; background:red; filter:alpha(opacity=50); -moz-opacity:0.5; -khtml-opacity:0.5; opacity:0.5'></div>
  <div id='zoomSensitiveZone' style='position:absolute; overflow:none; left:0px; top:0px; width:0px; height:0px; visibility:visible; cursor:crosshair; background:blue; filter:alpha(opacity=0); -moz-opacity:0; -khtml-opacity:0;opacity:0' oncontextmenu='return false' ></div>
<STYLE MEDIA="print">
        /*Turn off the zoomBox*/
        div#zoomBox, div#zoomSensitiveZone {display: none}
        /*This keeps IE from cutting things off*/
        #why {position: static; width: auto}
</STYLE>

DIV
          ;
        print "<tr align='center'><td><strong>Zoom graph => Start: "
          . scalar localtime( $info{graph_start} )
          . ", End: "
          . scalar localtime( $info{graph_end} )
          . " , graph Height=$graph_height Width=$graph_width</strong><br></td></tr>\n";

        print "<tr><td align='center'>\n";
        my $zoom_start = 2 * $info{graph_start} - $info{graph_end};
        my $zoom_end   = 2 * $info{graph_end} - $info{graph_start};
        print
"<a href='/$config{CGIBIN}/rrd2graph.cgi?action=zoom&graph_start=$zoom_start&graph_end=$info{graph_start}&graph_height=$graph_height&graph_width=$graph_width&hostname=$info{hostname}&service=$info{service}&title_font_size=10&el=$info{el}"
          . ( $info{ds_name} ? "&ds_name=$info{ds_name}'>" : "'>" )
          . " << Previous</a> | ";
        print
"<a href='/$config{CGIBIN}/rrd2graph.cgi?action=zoom&graph_start=$info{graph_end}&graph_end=$zoom_end&graph_height=$graph_height&graph_width=$graph_width&hostname=$info{hostname}&service=$info{service}&title_font_size=10&el=$info{el}"
          . ( $info{ds_name} ? "&ds_name=$info{ds_name}'>" : "'>" )
          . "Next >> </a>";
        print "</td></tr>\n";

        print "<tr><td>\n";
        if ( $info{ds_name} ) {
            print <<DSIMG;
	<img id='zoomGraphImage' src='/$config{CGIBIN}/rrd2graph.cgi?action=image&graph_start=$info{graph_start}&graph_end=$info{graph_end}&graph_height=$graph_height&graph_width=$graph_width&hostname=$info{hostname}&service=$info{service}&title_font_size=10&el=$info{el}&ds_name=$info{ds_name}' border='0' alt='$info{hostname} - $info{service}'> 
DSIMG
        }
        else {
            print <<IMG;
	<img id='zoomGraphImage' src='/$config{CGIBIN}/rrd2graph.cgi?action=image&graph_start=$info{graph_start}&graph_end=$info{graph_end}&graph_height=$graph_height&graph_width=$graph_width&hostname=$info{hostname}&service=$info{service}&title_font_size=10&el=$info{el}' border='0' alt='$info{hostname} - $info{service}'> 
IMG
        }
        print "</td></tr><br>\n";
        print "<tr align='center'><td>";
        print
"[<a href=\"/$config{CGIBIN}/rrd2graph.cgi?hostname=$info{hostname}&service=$info{service}&el=$info{el}"
          . ( $info{ds_name} ? "&ds_name=$info{ds_name}\">" : "\">" )
          . "Back</a>]";
        print " [<a href=\"javascript:help(zoom_help)\">Help</a>]";
        print $help_text;
        print "</td><tr></table> \n";

        #
        # load zoom java script
        open JS, "< $config{CONF_DIR}/$config{ZOOM_JS}"
          or log_msg
          "Can't open zoom script \"$config{CONF_DIR}/$config{ZOOM_JS}\"\n"
          and exit 1;

        while (<JS>) {
            s/\@CGIBIN\@/$config{CGIBIN}/;
            push @zoom_javascript, $_;
        }
        close JS;
        print @zoom_javascript;
        print "$footer" . $total_graphs;
        print " <br></body>\n\n";
        print "</html>\n\n";
    }
    elsif ( $info{action} eq "image" ) {

        foreach (@rrdtool_options) {
            s/"//g;
            s/\$DATA_FLOW\$/$start/g if $data_flow;
        }

        my ( $fh, $filename );
        $filename = '-';

        if ( $ENV{MOD_PERL} ) {    # This variable is set by mod_perl
            use CGI qw/:standard :imagemap/;
            print header('image/png');
            use File::Temp qw/ tempfile /;
            ( $fh, $filename ) = tempfile(
                DIR    => $config{TMPDIR},
                SUFFIX => '.image',
            );
        }
        else {
            print "Content-Type: image/png\n";
            print "\n";
        }

        if ( $RRDs::VERSION >= 1.2013 ) {
            push @rrdtool_options,
                "--watermark="
              . scalar localtime( $info{graph_start} ) . " to "
              . scalar localtime( $info{graph_end} );
        }

        my ( $a, $xsize, $ysize ) =
          RRDs::graph( $filename, "--start=$info{graph_start}",
            "--end=$info{graph_end}", @rrdtool_options );

        #
        # This work around is needed as RRDs::graph writes to STDOUT
        # under mod_perl the data is not flushed properly
        #
        if ( $ENV{MOD_PERL} ) {
            binmode $fh;
            my $buf;
            while ( sysread $fh, $buf, 8192 ) {
                print $buf;
            }
            close $fh;
            unlink $filename;
        }

        $RRD_ERROR = RRDs::error;
        log_msg
          "ERROR \"hostname = $info{hostname} -> service = $info{service}\"\
				 graph failed : $RRD_ERROR\n" if $RRD_ERROR;
    }
    else {
        log_msg "zoom: unknown type \"$info{action}\"\n";
    }
}

sub load_config ($) {
    my $config_file = shift;
    my $line        = 0;

    open( CONF, "<$config_file" )
      or die "Can't open configuration file -> \"$config_file\":$!\n";

    #
    # Parse configuration file
    while (<CONF>) {
        $line++;    # note confirguration file line number
        chomp;      # Remove newline character
        s/ //g;     # remove spaces
        next if /^#/;    # Skip comments
        next if /^$/;    # Skip empty lines
        s/#.*//;         # Remove partial comments

        if (/^(.*)=(.*)$/) {
            if ( defined $config{$1} ) {
                if ( $1 eq "THUMB_DISPLAY_COLUMNS" ) {
                    if ( ( $2 == 2 ) or ( $2 == 3 ) ) {
                        $config{$1} = $2;
                    }
                    else {
                        log_msg
"Wrong value supplied to \"THUMB_DISPLAY_COLUMNS\" must be 2 or 3\n";
                    }
                }
                else {
                    $config{$1} = $2;
                }
            }
            else {    # A warning should be ok
                log_msg
"WARNING: Unknown global variable declaration in line $line:\"$1 = $2\"\n";
            }
        }
    }
    close CONF;
}

sub load_nagios_env ($) {
    my $env_file = shift;

    my @env_choice = ();
    my %tmp        = ();

    #
    # read environment choice from template
    #
    my $env_template = "$config{CONF_DIR}/$config{TEMPLATES_DIR}/environment.t";
    open( E_TEMP, "<$env_template" )
      or die "Can't open Environment template file \"$env_template\"";
    while (<E_TEMP>) {
        chomp;     # Remove newline character
        s/ //g;    # remove spaces
        next if /^#/;               # Skip comments
        next if /^$/;               # Skip empty lines
        next if $_ !~ /^NAGIOS/;    # ignore junk
        s/#.*//;                    # Remove partial comments

        push @env_choice, $_;
    }
    log_msg "Environment template \"$env_template\" has no valid value\n"
      and exit 1
      if $#env_choice < 0;

    open( E, "<$env_file" )
      or die "Can't open Environment file -> \"$env_file\":$!\n";

    #
    # load environment variables
    while (<E>) {
        chomp;    # Remove newline character

        my ( $k, $v ) = split /=/, $_, 2;

        $v = localtime($v)
          if $k =~ /(LASTHOST|LASTSERVICE|PROCESSSTARTTIME)/;
        $tmp{$k} = $v;
    }
    close E;

    $nagios_env = "<table align='left' cellpadding='0'>\n";
    $nagios_env .=
"<br><tr><td bgcolor=\"gray\">MACRO NAME</td><td bgcolor=\"gray\">Value</td><tr>\n";
    foreach my $i (@env_choice) {
        my $result_str_color = "lightgrey";
        $result_str_color = 'red'        if $tmp{$i} =~ m/^(DOWN|CRITICAL)/i;
        $result_str_color = 'yellow'     if $tmp{$i} =~ m/^WARNING/i;
        $result_str_color = 'lightgreen' if $tmp{$i} =~ m/^(UP|OK)/i;
        $nagios_env .=
"<tr><td bgcolor=#ffe0dd>$i</td><td nowrap bgcolor=$result_str_color>$tmp{$i}</td></tr>\n"
          if defined $tmp{$i};
    }
    $nagios_env =~ s/NAGIOS_//g;
    $nagios_env .= "</table>\n";
}

sub list_default_values() {
    load_config($conf_file) if $conf_file;

    exit 0 if $ENV{'REQUEST_METHOD'};
    printf "#\n# Generated by rrd2graph: $version\n";
    printf "#%24s = %s\n", "Key  ", "  Value";
    printf "%25s = %s\n", "#" x 25, "#" x 30;
    foreach my $k ( sort keys %config ) {
        printf "%25s = %s\n", $k, $config{$k};
    }
    exit 0;
}

exit 0;
